/* Startup code compliant to the ELF Map ABI.
   Copyright (C) 1995, 1997, 2000, 2001 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <stdlib.h>

/* This is the canonical entry point, usually the first thing in the text
   segment.  The SVR4/Mips ABI (pages 3-31, 3-32) says that when the entry
   point runs, most registers' values are unspecified, except for:

   v0 ($2)	Contains a function pointer to be registered with `atexit'.
		This is how the dynamic linker arranges to have DT_FINI
		functions called for shared libraries that have been loaded
		before this code runs.

   sp ($29)	The stack contains the arguments and environment:
		0(%esp)			argc
		4(%esp)			argv[0]
		...
		(4*argc)(%esp)		NULL
		(4*(argc+1))(%esp)	envp[0]
		...
					NULL
   ra ($31)	The return address register is set to zero so that programs
		that search backword through stack frames recognize the last
		stack frame.
*/


/* We need to call:
   __libc_start_main (int (*main) (int, char **, char **), int argc,
		      char **argv, void (*init) (void), void (*fini) (void),
		      void (*rtld_fini) (void), void *stack_end)
*/

extern int main(int, char**, char**);

extern void _init(void);
extern void _fini(void);

extern int __libc_start_main (int (*main) (int, char **, char **), int argc,
      char **argv, void (*init) (void), void (*fini) (void),
      void (*rtld_fini) (void), void *stack_end);

extern void __eti_register_atomic(void);
extern int  __syscall_eti_signal_return(void* fun);
extern void __eti_signal_return_handler();

void _start(int argc, char** argv)
{
   void* low_sp = (void*) ((unsigned long) &low_sp & 31);

   __eti_register_atomic();
   __syscall_eti_signal_return((void*) __eti_signal_return_handler);

   __libc_start_main(main, argc, argv, _init, _fini, NULL, low_sp);

   /*
    * Dump core if we ever return here.
    */
   *(int*) 0;
}
